@@ -25,8 +25,11 @@ from docassemble.base.util import (
     space_to_underscore,
     DAStaticFile,
     alpha,
+    showifdef,
 )
 from docassemble.base.pdfa import pdf_to_pdfa
+from textwrap import wrap
+from math import floor
 
 __all__ = [
     "ALAddendumField",
@@ -141,10 +144,18 @@ class ALAddendumField(DAObject):
         preserve_newlines: bool = False,
         input_width: int = 80,
         overflow_message: str = "",
+        preserve_words: bool = True,
     ):
         """
         Try to return just the portion of the variable (list-like object or string)
         that is not contained in the safe_value().
+
+        Whitespace will be altered. If preserve_newlines is true, the return value may have newlines,
+        but double newlines and Windows style (\r\n) will be replaced with \n. Double spaces will replaced
+        with a single space.
+
+        If preserve_newlines is false, all whitespace, including newlines and tabs, will be replaced
+        with a single space.
         """
         # Handle a Boolean overflow first
         if isinstance(self.overflow_trigger, bool) and self.overflow_trigger:
@@ -157,12 +168,20 @@ class ALAddendumField(DAObject):
             input_width=input_width,
             preserve_newlines=preserve_newlines,
             _original_value=original_value,
+            preserve_words=preserve_words,
         )
         if isinstance(safe_text, str):
-            # Always get rid of double newlines, for consistency with safe_value.
-            value_to_process = re.sub(
-                r"[\r\n]+|\r+|\n+", r"\n", original_value
-            ).rstrip()
+            max_lines = self.max_lines(input_width=input_width)
+
+            if preserve_newlines and max_lines > 1:
+                # we do our own substitution of whitespace, only double newlines and spaces
+                value_to_process = (
+                    re.sub(r"[\r\n]+|\r+|\n+", r"\n", original_value).replace("  ", " ")
+                ).rstrip()
+            else:
+                # textwrap.wrap(replace_whitespace=True) replaces all whitespace, not just double newlines and spaces
+                value_to_process = re.sub(r"\s+", " ", original_value).strip()
+
             if safe_text == value_to_process:  # no overflow
                 return ""
             # If this is a string, the safe value will include an overflow message. Delete
@@ -174,19 +193,16 @@ class ALAddendumField(DAObject):
             #   1. We replace all double newlines with \n.
             #   2. Character count will adjust to reflect double-newlines being replaced with one char.
             overflow_start = max(len(safe_text) - len(overflow_message), 0)
-            return value_to_process[overflow_start:]
+            return value_to_process[overflow_start:].lstrip()
 
         # Do not subtract length of overflow message if this is a list of objects instead of a string
         return original_value[self.overflow_trigger :]
 
-    def max_lines(self, input_width: int = 80, overflow_message_length=0) -> int:
+    def max_lines(self, input_width: int = 80) -> int:
         """
-        Estimate the number of rows in the field in the output document.
+        Calculate the number of lines of text that will fit in the specified input
         """
-        return (
-            int(max(self.overflow_trigger - overflow_message_length, 0) / input_width)
-            + 1
-        )
+        return floor(self.overflow_trigger / input_width)
 
     def value(self) -> Any:
         """
@@ -201,11 +217,24 @@ class ALAddendumField(DAObject):
         overflow_message: str = "",
         input_width: int = 80,
         preserve_newlines: bool = False,
-        _original_value=None,
-    ):
+        _original_value: Optional[str] = None,
+        preserve_words: bool = True,
+    ) -> Union[str, List[Any]]:
         """
-        Try to return just the portion of the variable
-        that is _shorter than_ the overflow trigger. Otherwise, return empty string.
+        Return just the portion of the variable that heuristics suggest will fit in the specified overflow_trigger
+        limit. If the value is not defined, return empty string.
+
+        When `preserve_newlines` is `True`, the output will be limited to a number of lines, not a number
+        of characters. The max number of lines will be calculated as `floor(self.overflow_trigger/input_width)`.
+        Therefore, it is important that `input_width` is a divisor of `overflow_trigger`.
+
+        Whitespace will be altered. If preserve_newlines is true, the return value may have newlines,
+        but double newlines and Windows style (\r\n) will be replaced with \n. Double spaces will replaced
+        with a single space.
+
+        If preserve_newlines is false, all whitespace, including newlines and tabs, will be replaced
+        with a single space.
+
         Args:
             overflow_message (str): A short message to go on the page where text is cutoff.
             input_width (int): The width, in characters, of the input box. Defaults to 80.
@@ -229,54 +258,43 @@ class ALAddendumField(DAObject):
         ):
             return value
 
-        max_lines = self.max_lines(
-            input_width=input_width, overflow_message_length=len(overflow_message)
-        )
+        max_lines = self.max_lines(input_width=input_width)
         max_chars = max(self.overflow_trigger - len(overflow_message), 0)
 
-        # If there are at least 2 lines, we can ignore overflow trigger.
-        # each line will be at least input_width wide
-        if preserve_newlines and max_lines > 1:
-            if isinstance(value, str):
+        # Strip newlines from strings because they take extra space
+        if isinstance(value, str):
+            # If we preserve newlines, we need to account for max_lines, not just max_chars
+            if preserve_newlines and max_lines > 1:
                 # Replace all new line characters with just \n. \r\n inserts two lines in a PDF
                 value = re.sub(r"[\r\n]+|\r+|\n+", r"\n", value).rstrip()
-                line = 1
-                retval = ""
-                paras = value.split("\n")
-                para = 0
-                while line <= max_lines and para < len(paras):
-                    # add the whole paragraph if less than width of input
-                    if len(paras[para]) <= input_width:
-                        retval += paras[para] + "\n"
-                        line += 1
-                        para += 1
-                    else:
-                        # Keep taking the first input_width characters until we hit max_lines
-                        # or we finish the paragraph
-                        while line <= max_lines and len(paras[para]):
-                            retval += paras[para][:input_width]
-                            paras[para] = paras[para][input_width:]
-                            line += 1
-                        if not len(paras[para]):
-                            para += 1
-                            retval += "\n"
-                # TODO: check logic here to only add overflow message when we exceed length
-                if len(paras) > para:
-                    return (
-                        retval.rstrip() + overflow_message
-                    )  # remove trailing newline before adding overflow message
-                else:
-                    return retval
-
-        # Strip newlines from strings
-        if isinstance(value, str):
-            if len(value) > self.overflow_trigger:
+                # textwrap.wrap does all the hard work for us here
                 return (
-                    re.sub(r"[\r\n]+|\r+|\n+", " ", value).rstrip()[:max_chars]
-                    + overflow_message
+                    " ".join(
+                        wrap(
+                            value,
+                            width=input_width,
+                            max_lines=max_lines,
+                            replace_whitespace=False,
+                            placeholder=overflow_message,
+                        )
+                    )
+                    .replace("  ", " ")
+                    .rstrip()
                 )
+
+            value = re.sub(r"\s+", " ", value)
+            if len(value) > self.overflow_trigger:
+                if preserve_words:
+                    retval = wrap(
+                        value,
+                        width=max_chars,
+                        replace_whitespace=True,
+                        drop_whitespace=True,
+                    )
+                    return next(iter(retval)).rstrip() + overflow_message
+                return value.rstrip()[:max_chars] + overflow_message
             else:
-                return re.sub(r"[\r\n]+|\r+|\n+", " ", value).rstrip()[:max_chars]
+                return value.rstrip()
 
         # If the overflow item is a list or DAList
         if isinstance(value, list) or isinstance(value, DAList):
@@ -290,9 +308,7 @@ class ALAddendumField(DAObject):
         Return the value of the field if it is defined, otherwise return an empty string.
         Addendum should never trigger docassemble's variable gathering.
         """
-        if defined(self.field_name):
-            return value(self.field_name)
-        return ""
+        return showifdef(self.field_name, "")
 
     def __str__(self):
         return str(self.value_if_defined())
@@ -798,6 +814,7 @@ class ALDocument(DADict):
         overflow_message: Optional[str] = None,
         preserve_newlines: bool = False,
         input_width: int = 80,
+        preserve_words: bool = True,
     ):
         """
         Shortcut syntax for accessing the "safe" (shorter than overflow trigger)
@@ -809,6 +826,7 @@ class ALDocument(DADict):
             overflow_message=overflow_message,
             preserve_newlines=preserve_newlines,
             input_width=input_width,
+            preserve_words=preserve_words,
         )
 
     def overflow_value(
@@ -817,6 +835,7 @@ class ALDocument(DADict):
         overflow_message: Optional[str] = None,
         preserve_newlines: bool = False,
         input_width: int = 80,
+        preserve_words: bool = True,
     ):
         """
         Shortcut syntax for accessing the "overflow" value (amount that exceeds overflow trigger)
@@ -830,6 +849,7 @@ class ALDocument(DADict):
             overflow_message=overflow_message,
             preserve_newlines=preserve_newlines,
             input_width=input_width,
+            preserve_words=preserve_words,
         )
 
     def is_enabled(self, refresh=True):
